"""
Cell module for Metamaterial Analysis Code (MAC). It represents a single cell, made by different elements.
"""

from .MACNode import MACNode
from .MACElement import MACElement
from .MACProperty import MACProperty
from .MACMaterial import MACMaterial
from .MACStructure import MACAuxetic

class MACCell:
    """
    Cell class for Metamaterial Analysis Code (MAC). It represents a single cell, made by different elements.

    Attributes:
        ID: cell ID
        Type: type of the cell (BBC, FCC, Auxetic, etc)
        Coords: Coordinates of the center of the cell
        Size: Size of the cell
        Elements: list of elements that compose the cell
        Nodes: list of nodes that compose the cell
        Material: list of the materials of the cell
        Properties: list of the properties of the cell

    """

    def __init__(self, id: int, structure: MACAuxetic, coords: tuple, elements: tuple[MACElement], nodes: tuple[MACNode],
                 material: list[MACMaterial], properties: list[MACProperty]):
        """
        Constructor for MACCell class
        """
        self.__id = id
        self.__structure = structure
        self.__coords = coords
        self.__elements = elements
        self.__nodes = nodes
        self.__material = material
        self.__properties = properties

    @property
    def ID(self) -> int:
        return self.__id

    @property
    def Structure(self) -> str:
        return self.__structure

    @property
    def Coords(self) -> tuple:
        return self.__coords

    @property
    def Size(self) -> float:
        return self.__size

    @property
    def Elements(self) -> tuple:
        return self.__elements

    @property
    def Nodes(self) -> tuple:
        return self.__nodes

    @property
    def Material(self) -> list:
        return self.__material

    @property
    def Properties(self) -> list:
        return self.__properties

    @Elements.setter
    def Elements(self, elements) -> None:
        self.__elements = elements




"""
FEM model module for the Metamaterial Analysis Code (MAC) package. It represents the whole model.
"""
from datetime import datetime
from typing import TypedDict

from .MACMaterial import MACMaterial
from .MACProperty import MACProperty
from .MACStructure import MACAuxetic
from .MACGlobals import MAC_VERSION, ELEMENTS_SET, NODES_DICT
from .MACCell import MACCell
from MetamaterialAnalysisCode._modules._centers_RPA import _centers_RPA
from MetamaterialAnalysisCode.classes.MACElement import MACElement


class MACModel:
    """
    FEM model class for the Metamaterial Analysis Code (MAC) package. It represents the whole model.

    Attributes:
        ModelDimensions: dimensions of the model
        CellStructure: structure of the cell (BBC, FCC, Auxetic, etc)
        CellSize: size of the cell
        CellMaterial: MACMaterial of the cell
        CellProperty: MACProperty of the cell
    """

    def __init__(self, modeldimensions: tuple[float], cellstructure: MACAuxetic, cellmaterial: list[MACMaterial],
                 cellproperty: list[MACProperty]):
        """
        Constructor for MACModel class
        """
        self.__modeldimensions = modeldimensions
        self.__cellstructure = cellstructure
        self.__cellmaterial = cellmaterial
        self.__cellproperty = cellproperty
        self.__celldict = {}

        # Function to generate the cells centers for auxetic structures
        cloud_set = _centers_RPA(a=round((3**0.5) * 0.5 * self.__cellstructure.Stepy, 4),
                                 h=round(self.__cellstructure.Hprisma, 4),
                                 stx=round(self.__cellstructure.Stepx, 4),
                                 dimensions=self.__modeldimensions
                                 )

        for i, center in enumerate(cloud_set):
            nodes, elements = self.__cellstructure.build(
                                                         cellcenter=center, mat=self.__cellmaterial[0],
                                                         prop=self.__cellproperty[0], vvector=(0.0, 0.0, 1.0)
                                                        )

            self.__celldict[i] = MACCell(i, self.__cellstructure, center, elements, nodes, self.__cellmaterial,
                                         self.__cellproperty)

        self.NodeDict = {node.ID: node for node in NODES_DICT.values()}
        self.ElementDict: ElementDictType = {element.ID: element for element in ELEMENTS_SET}

    def write_fem(self, path: str, writeheader: bool = True) -> None:
        """
        Writes the .fem file for the FEM model in the specified path. It appends the data to the file if it already
        exists.
        """

        if path[-4:] != ".fem":
            path += ".fem"

        with open(path, "a") as w:

            if writeheader:

                today = datetime.now().strftime("%d/%m/%Y - %H:%M:%S")
                analysis_name = path.split("\\")[-1].split(".")[0]
                macverlen = 78 - (len(MAC_VERSION) + len("$$  MAC version: "))
                analysis_namelen = 78 - (len(analysis_name) + len("$$  Analysis Name: "))

                header = "$" * 80 + "\n" + \
                         "$" * 13 + "  File generated by Metamaterial Analysis Code (MAC)  " + "$" * 13 + "\n" + \
                         "$" * 80 + "\n" + \
                         "$$" + " " * 76 + "$$" + "\n" + \
                         "$$  Escuela Tecnica Superior de Ingenieria Aeronautica y del Espacio" + " " * 10 + "$$" + "\n" + \
                         "$$" + " " * 76 + "$$" + "\n" + \
                         f"$$  MAC version: {MAC_VERSION}" + " " * macverlen + "$$" + "\n" + \
                         "$$" + " " * 76 + "$$" + "\n" + \
                         f"$$  Date: {today}" + " " * 47 + "$$" "\n" + \
                         "$$" + " " * 76 + "$$" + "\n" + \
                         f"$$  Analysis Name: {analysis_name}" + " " * analysis_namelen + "$$" + "\n" + \
                         "$$" + " " * 76 + "$$" + "\n" + \
                         "$" * 80 + "\n"
                w.writelines(header)

            w.writelines("\n$$\n$$  GRID Data\n$$\n")

            for node in sorted(self.NodeDict.keys()):
                w.writelines(str(self.NodeDict[node]))

            w.writelines("\n$$\n$$  Elements\n$$\n")

            for element in self.ElementDict.values():
                    w.writelines(str(element))

            w.writelines("\n$$\n$$ Properties\n$$\n")

            for prop in self.__cellproperty:
                w.writelines(str(prop))

            w.writelines("\n$$\n$$ Materials\n$$\n")

            for material in self.__cellmaterial:
                w.writelines(str(material))

    def __del__(self):
        """
        Destructor for MACModel class
        """
        ELEMENTS_SET.clear()
        NODES_DICT.clear()


def set_model(**kwargs) -> MACModel:
    """
    Function that creates a MACModel object. It uses the kwargs dictionary.
    model = set_model(modeldimensions=(x, y, z), cellstructure=MACAuxetic, cellmaterial=[MACMaterial],
                      cellproperty=[MACProperty])
    """
    model = MACModel(modeldimensions=kwargs["modeldimensions"], cellstructure=kwargs["cellstructure"],
                     cellmaterial=kwargs["cellmaterial"], cellproperty=kwargs["cellproperty"])
    return model


class ElementDictType(TypedDict):
    """
    Dictionary with the elements of the model. The key is the element ID and the value is a MACElement.
    """
    int: MACElement




"""
This is a hidden module that creates the Right Prism Auxetic cell centers.
"""

import numpy as np


def _centers_RPA(a, h, stx, dimensions) -> np.ndarray[np.ndarray[np.float16, np.float16]]:
    """
    Function that creates the cloud of the center points of the Auxetic cells that shape the Right Prism model.
    """
    ncx = int(dimensions[0]//(1.5*stx))
    ncy = int(dimensions[1]//(2*a))
    ncz = int(dimensions[2]//h)

    n = ncx*ncy*ncz
    centers = np.zeros((n, 3))

    centers[0] = np.array([stx, a, round(h/2, 4)])

    for i in range(ncx):
        for j in range(ncy):
            for k in range(ncz):
                if i == 0 and j == 0 and k == 0:
                    continue
                elif i % 2 == 0:
                    centers[i*ncy*ncz + j*ncz + k] = np.array([stx + i*1.5*stx, a + j*2*a, k*h + round(h/2, 4)])
                else:
                    centers[i*ncy*ncz + j*ncz + k] = np.array([stx + i*1.5*stx, 2*a + j*2*a, k*h + round(h/2, 4)])

    return centers



"""
Property module with properties classes for the Metamaterial Analysis Code (MAC). There are one class per type of
property. They all derivative of the base class MACProperty, which is an abstract class.
"""
from abc import ABC, abstractmethod

from .MACMaterial import MACMaterial


class MACProperty(ABC):
    """
    Property class for the Metamaterial Analysis Code (MAC). It represents a single property.

    Attributes:
        ID: property ID
        Type: type of the property (PBEAM, PSOLID, etc)
        Material: list of the materials of the property
    """

    @abstractmethod
    def __init__(self, id: int, type: str, material: list[MACMaterial]):
        """
        Constructor for MACProperty class
        """
        self.__id = id
        self.__type = type
        self.__material = material

    @property
    def ID(self):
        return self.__id

    @property
    def Type(self):
        return self.__type

    @property
    def Material(self):
        return self.__material


class MACBeam(MACProperty):
    """
    Beam property class for the Metamaterial Analysis Code (MAC). It represents a single beam property.

    Attributes:
        Area: area of the beam
        I1: Area moment of inertia of the beam
        I2: Area moment of inertia of the beam
        I12: area product of inertia of the beam
        J: torsional stiffness of the beam
    """

    def __init__(self, id: int, type: str, material: list, area: float, i1: float, i2: float, i12: float, j: float):
        """
        Constructor for MACBeam class
        """
        super().__init__(id, type, material)
        self.__area = area
        self.__i1 = i1
        self.__i2 = i2
        self.__i12 = i12
        self.__j = j

    @property
    def Area(self):
        return self.__area

    @property
    def I1(self):
        return self.__i1

    @property
    def I2(self):
        return self.__i2

    @property
    def I12(self):
        return self.__i12

    @property
    def J(self):
        return self.__j

    # Method to print the property. It uses the 8 characters format of Optistruct.
    def __str__(self):

        idspaces = " " * (8 - len(str(self.ID)))
        materialspaces = " " * (8 - len(str(self.Material[0].ID)))
        area = "{:.5f}".format(self.Area)[:7]
        inertia1 = "{:.5f}".format(self.I1)[:7]
        inertia2 = "{:.5f}".format(self.I2)[:7]
        inertia12 = "{:.5f}".format(self.I12)[:7]
        torsion = "{:.5f}".format(self.J)[:7]

        return f"PBEAM   {self.ID}{idspaces}{self.Material[0].ID}{materialspaces}{area} {inertia1} {inertia2} " + \
               f"{inertia12} {torsion} \n"


class MACBeamL(MACProperty):
    """
    Beam property with a specific section. Based on the PBEAML property of optistruct.

    Attributes:
        Section: Type of section. Supported: ROD
        Dim1: First dimension. A value must be given
    """
    def __init__(self, id: int, type: str, material: list[MACMaterial], section: str, dim1: float):
        super().__init__(id, type, material)
        self.Section = section
        self.Dim1 = dim1

    # Method to print the property. It uses the 8 characters format of Optistruct.
    def __str__(self):

        idspaces = " " * (8 - len(str(self.ID)))
        materialspaces = " " * (8 - len(str(self.Material[0].ID)))
        eight = " "*8
        sectspaces = " " * (8 - len(str(self.Section)))
        dim1 = "{:.5f}".format(self.Dim1)[:7]

        return f"PBEAML  {self.ID}{idspaces}{self.Material[0].ID}{materialspaces}{eight}{self.Section}{sectspaces}\n" +\
               f"{eight}{dim1}\n"


def set_property(**kwargs) -> MACBeam | MACBeamL:
    """
    Function to create a MACProperty based object. It uses the kwargs dictionary. Supported subclasses are:
        - MACBeam:  beam = set_property(id=int, type="PBEAM", material=[MACMaterial], area=float, i1=float, i2=float, i12=float, j=float)
        - MACBeamL: beam = set_property(id=int, type="PBEAML", material=[MACMaterial], section="ROD", dim1=float)
    """

    if kwargs["type"] == "PBEAM":
        return MACBeam(id=kwargs["id"], type=kwargs["type"], material=kwargs["material"], area=kwargs["area"],
                       i1=kwargs["i1"], i2=kwargs["i2"], i12=kwargs["i12"], j=kwargs["j"])

    elif kwargs["type"] == "PBEAML":

        return MACBeamL(id=kwargs["id"], type=kwargs["type"], material=kwargs["material"], section=kwargs["section"],
                        dim1=kwargs["dim1"])

    else:
        raise ValueError("The property type is not supported.")


"""
Module with the MACSubcase class
"""

from MetamaterialAnalysisCode.classes.MACLoadCase import MACForce
from MetamaterialAnalysisCode.classes.MACLoadCase import MACSpc
from MetamaterialAnalysisCode.classes.MACLoadCase import MACEigrl
from MetamaterialAnalysisCode.classes.MACNonLinear import MACNLout, MACNLparmLD


class MACSubcase:
    """
    Class for the subcase.

    Attributes:
        ID: subcase ID
        Label: subcase label
        LoadCases: list of load cases of the subcase
        StatSub: linear static subcase. Some subcases(buckling) need a linear static subcase to work
    """

    def __init__(self, id_: int, label: str, load: list[MACForce | MACSpc, ...], spc: list[MACSpc, ...],
                 eigr: MACEigrl = None, stat_sub: 'MACSubcase' = None, nlaprmld: MACNLparmLD = None,
                 nlout: MACNLout = None, output: list[str] = None):
        """
        Constructor for MACSubcase class
        """
        ids = set()
        for i in load:
            ids.add(i.ID)
        if len(ids) > 1:
            raise ValueError("All the loads of a subcase must have the same ID")

        ids = set()
        for i in spc:
            ids.add(i.ID)
        if len(ids) > 1:
            raise ValueError("All the constrains of a subcase must have the same ID")

        self.ID = id_
        self.Label = label
        self.Load = load
        self.Spc = spc
        self.Eigr = eigr
        self.StatSub = stat_sub
        self.NLaprmLD = nlaprmld
        self.NLout = nlout
        self.Output = output

    def __str__(self):
        """
        Method to print a subcase. It uses the 8 characters format of Optistruct.
        """

        idspaces = " " * 8
        case = ""
        outstr = ""
        if isinstance(self.Output, list):
            for out in self.Output:
                outstr += "  " + out + "\n"

        if not self.Label.startswith("buckling") and isinstance(self.Load[0], MACForce):
            loadspaces = " " * (8 - len(str(self.Load[0].ID)))
            case += f"  LOAD = {loadspaces}{self.Load[0].ID}\n"

        elif not self.Label.startswith("buckling") and isinstance(self.Load[0], MACSpc):
            loadspaces = " " * (8 - len(str(self.Load[0].ID)))
            case += f"  LOAD = {loadspaces}{self.Load[0].ID}\n"

        loadspaces = " " * (8 - len(str(self.Spc[0].ID)))
        case += f"  SPC = {loadspaces}{self.Spc[0].ID}\n"

        # returns the subcase depending on its label -------------------------------------------------------------------
        if self.Label.startswith("linear"):
            return f"SUBCASE{idspaces}{self.ID}\n" + f"  LABEL {self.Label}\n" + case + outstr

        elif self.Label.startswith("buckling"):
            if self.StatSub is None:
                raise ValueError("Buckling subcase needs a linear static subcase")
            else:
                subspaces = " " * 8
                methodspace = " " * 8

            return f"SUBCASE{idspaces}{self.ID}\n" + f"  LABEL {self.Label}\n" + "ANALYSIS BUCK\n" + case + \
                   f"  METHOD(STRUCTURE) ={methodspace}{self.Eigr.ID}\n" + \
                   f"  STATSUB(BUCKLING) ={subspaces}{self.StatSub.ID}\n" + outstr

        elif self.Label.startswith("nonlinear"):
            return f"SUBCASE{idspaces}{self.ID}\n" + f"  LABEL {self.Label}\n" + "ANALYSIS NLSTAT\n" + case + \
                   f"  NLPARM(LGDISP) =        {self.NLaprmLD.ID}\n" + \
                   f"  NLOUT =        {self.NLout.ID}\n" + outstr


def set_subcase(id: int, label: str, constraints: list[MACSpc, ...], loads: list[MACForce | MACSpc, ...] = None,
                eigr: MACEigrl = None, stat_sub: object = None, nlparmld: MACNLparmLD = None, nlout: MACNLout = None,
                output: list[str] = None):
    """
    Function to set a subcase. It returns a MACSubcase object.

    Arguments:
        - id: subcase ID
        - label: subcase label ("linear", "buckling")
        - load_cases: list of load cases of the subcase
        - stat_sub: (opcional) linear static subcase. Some subcases (buckling) need a linear static subcase to work
    """
    if loads is None:
        loads = []

    # check if all the loads have the same ID
    ids = set()
    for i in loads:
        ids.add(i.ID)
    if len(ids) > 1:
        raise ValueError("All the loads of a subcase must have the same ID")

    # check if all the constraints have the same ID
    ids = set()
    for i in constraints:
        ids.add(i.ID)
    if len(ids) > 1:
        raise ValueError("All the constrains of a subcase must have the same ID")

    return MACSubcase(id, label, loads, constraints, eigr, stat_sub, nlparmld, nlout, output)



"""
Main module for the Metamaterial Analysis Code (MAC) package. It represents the whole analysis.
"""

from os import remove, path
from datetime import datetime

from .MACModel import MACModel
from .MACGlobals import MAC_VERSION
from MetamaterialAnalysisCode.classes.MACSubcase import MACSubcase
from MetamaterialAnalysisCode.classes.MACLoadCase import MACForce, MACSpc, MACEigrl

class MACAnalysis:
    """
    Main class for the Metamaterial Analysis Code (MAC) package. It represents the whole analysis.

    Attributes:
        Model: MACModel of the model
        Subcases: list of MACSubcase
        Parameters: list of strings with the parameters.
    """

    def __init__(self, model: MACModel, subcases: list[MACSubcase, ...], parameters: list[str, ...]):
        """
        Constructor for MACAnalysis class
        """
        self.Model = model
        self.Subcases = subcases
        self.Parameters = parameters

    @property
    def Loads(self) -> list[MACForce | MACSpc, ...]:
        loads = []
        for subcase in self.Subcases:
            for load in subcase.Load:
                if load not in loads:
                    loads.append(load)

        return loads

    @property
    def Spcs(self) -> list[MACSpc, ...]:
        spcs = []
        for subcase in self.Subcases:
            for spc in subcase.Spc:
                if spc not in spcs:
                    spcs.append(spc)
        return spcs

    @property
    def Eigrs(self) -> list[MACEigrl, ...]:
        eigrs = []
        for subcase in self.Subcases:
            if subcase.Eigr is not None and subcase.Eigr not in eigrs:
                eigrs.append(subcase.Eigr)
        return eigrs

    def write_fem(self, pathfem: str) -> None:
        """
        Writes the .fem file for the analysis in the specified path. It overwrites the file if it already exists.
        """

        if pathfem[-4:] != ".fem":
            pathfem += ".fem"

        if path.exists(pathfem):
            remove(pathfem)
            print(f"File {pathfem} already exists. It has been overwritten.")

        today = datetime.now().strftime("%d/%m/%Y - %H:%M:%S")
        analysis_name = pathfem.split("\\")[-1].split(".")[0]
        macverlen = 78 - (len(MAC_VERSION) + len("$ MAC version: "))
        analysis_namelen = 78 - (len(analysis_name) + len("$ Analysis Name: "))

        header = "$" * 82 + "\n" + \
                 "$" * 14 + "  File generated by Metamaterial Analysis Code (MAC)  " + "$" * 14 + "\n" + \
                 "$" * 82 + "\n" + \
                 "$$" + " " * 78 + "$$" + "\n" + \
                 "$$  Escuela Tecnica Superior de Ingenieria Aeronautica" + " " * 26 + "$$" + "\n" + \
                 "$$" + " " * 78 + "$$" + "\n" + \
                 f"$$  MAC version: {MAC_VERSION}" + " " * macverlen + "$$" + "\n" + \
                 "$$" + " " * 78 + "$$" + "\n" + \
                 f"$$  Date: {today}" + " " * 49 + "$$" "\n" + \
                 "$$" + " " * 78 + "$$" + "\n" + \
                 f"$$  Analysis Name: {analysis_name}" + " " * analysis_namelen + "$$" + "\n" + \
                 "$$" + " " * 78 + "$$" + "\n" + \
                 "$" * 82 + "\n"

        with open(pathfem, "a") as w:
            w.writelines(header)
            w.writelines("\n$$" + "-"*78 + "$$\n" +
                         "$$                       Case Control Cards                                     $$\n" +
                         "$$" + "-"*78 + "$$\n\n")
            for subcase in self.Subcases:
                w.writelines(str(subcase))
                w.writelines("\n")
            if self.Parameters is not None:
                for parameter in self.Parameters:
                    w.writelines(parameter)
            w.writelines("\nBEGIN BULK\n")

        # Write model calling the model method. header=False to avoid writing the header again.
        self.Model.write_fem(pathfem, writeheader=False)

        with open(pathfem, "a") as w:

            w.writelines("\n$$\n$$ Load cases\n$$\n")

            for load in self.Loads:
                w.writelines(str(load))
            for spc in self.Spcs:
                w.writelines(str(spc))
            for eigr in self.Eigrs:
                if eigr is not None:
                    w.writelines("\n" + str(eigr))
            for subcase in self.Subcases:
                if subcase.NLaprmLD is not None:
                    w.writelines("\n" + str(subcase.NLaprmLD))
            for subcase in self.Subcases:
                if subcase.NLout is not None:
                    w.writelines("\n" + str(subcase.NLout))

            w.writelines("$$\n")
            w.writelines("\n\nENDDATA\n")


def set_analysis(model: MACModel, subcases: list, parameters: list = None) -> MACAnalysis:
    """
    Function to set the analysis. It returns a MACAnalysis object.
    """
    analysis = MACAnalysis(model, subcases, parameters)

    return analysis



"""
Module for the MACTable class. It represents a tables that optistruct uses for different purposes. For example, define
the strain-stress curve of a material.
"""


class MACTable:
    """
    MACTable class. It represents a tables that optistruct uses for different purposes. For example, define the
    strain-stress curve of a material.
    """

    def __init__(self, id_: int, data: list):
        """
        Constructor for MACTable class
        """
        self.ID = id_
        self.Data = data


def set_table(id: int, data: list) -> MACTable:
    """
    Function to create a MACTable object.
    """
    table = MACTable(id, data)
    return table



"""
Module for the NonLinear Parameters for Large Displacements (NLaprmLD) class and NonLinear Output (NLout) class.
"""


class MACNLparmLD:
    """
    Class for the NonLinear Parameters for Large Displacements (NLparmLD).
    """

    def __init__(self, id: int, dt: float):
        """
        Constructor for MACNLparmLD class
        """
        self.ID = id
        self.DT = dt

    def __str__(self):
        """
        Method to print the NonLinear Parameters for Large Displacements (NLaprmLD) class.
        """
        idspaces = " " * (8 - len(str(self.ID)))
        dtspaces = " " * (8 - len(str(self.DT)))
        return f"NLPARM  {idspaces}{self.ID}        {dtspaces}{self.DT}\n"


class MACNLout:
    """
    Class for the NonLinear Output (NLout).
    """

    def __init__(self, id: int, nint: int):
        """
        Constructor for MACNLout class
        """
        self.ID = id
        self.NINT = nint

    def __str__(self):
        """
        Method to print the NonLinear Output (NLout) class.
        """
        idspaces = " " * (8 - len(str(self.ID)))
        nintspaces = " " * (8 - len(str(self.NINT)))
        return f"NLOUT   {idspaces}{self.ID}    NINT{nintspaces}{self.NINT}\n"


def set_nlparmld(id: int, dt: float):
    """
    Function to create a NonLinear Parameters for Large Displacements (NLaprmLD) class.
    """
    return MACNLparmLD(id, dt)


def set_nlout(id: int, nint: int):
    """
    Function to create a NonLinear Output (NLout) class.
    """
    return MACNLout(id, nint)


"""
Cell Structure module for Metamaterial Analysis Code (MAC). It represents the structure.
"""
import sys

from .MACMaterial import MACMaterial
from .MACProperty import MACProperty
from .MACNode import MACNode
from .MACElement import MACElement
from .MACGlobals import NODES_DICT, ELEMENTS_SET


class MACAuxetic:
    """
    Auxetic Structure class for Metamaterial Analysis Code (MAC). It represents the auxetic structure.

    Attributes:
        Djoint: distance between the auxiliary nodes (M1, M2) and the horizontal central plane
        Dstar: distance between the first group of nodes (B1,...,B6) and the center of the cell
        Heightstar: distance between the second group of nodes (S1,...,S6) and the horizontal central plane
    """

    def __init__(self, djoint: float, dstar: float, heightstar: float, hcapas: float,hprisma: float,
                 stepx: float, stepy: float, nelem: int):
        """
        Constructor for MACAuxetic class
        """

        self.__djoint = djoint
        self.__dstar = dstar
        self.__heightstar = heightstar
        self.__hcapas = hcapas
        self.__hprisma = hprisma
        self.__stepx = stepx
        self.__stepy = stepy
        self.Type = "Auxetic"
        self.__nelem = nelem

    @property
    def Djoint(self) -> float:
        return self.__djoint

    @property
    def Dstar(self) -> float:
        return self.__dstar

    @property
    def Heightstar(self) -> float:
        return self.__heightstar

    @property
    def Hcapas(self) -> float:
        return self.__hcapas

    @property
    def Hprisma(self) -> float:
        return self.__hprisma

    @property
    def Stepx(self) -> float:
        return self.__stepx

    @property
    def Stepy(self) -> float:
        return self.__stepy

    def build(self, cellcenter: tuple[float, ...], mat: MACMaterial, prop: MACProperty, vvector: tuple)\
            -> tuple[tuple[MACNode, ...], tuple[MACElement, ...]]:

        # M1
        x = round(cellcenter[0], 4)
        y = round(cellcenter[1], 4)
        z = round(cellcenter[2] + self.__djoint, 4)
        M1 = NODES_DICT.get((x, y, z), MACNode(len(NODES_DICT) + 1, (x, y, z)))
        NODES_DICT[(x, y, z)] = M1

        # M2
        x = round(cellcenter[0], 4)
        y = round(cellcenter[1], 4)
        z = round(cellcenter[2] - self.__djoint, 4)
        M2 = NODES_DICT.get((x, y, z), MACNode(len(NODES_DICT) + 1, (x, y, z)))
        NODES_DICT[(x, y, z)] = M2

        # S1
        x1 = round(cellcenter[0] + self.__stepx, 4)
        y1 = round(cellcenter[1], 4)
        z1 = round(cellcenter[2] + self.__heightstar, 4)
        S1 = NODES_DICT.get((x1, y1, z1), MACNode(len(NODES_DICT) + 1, (x1, y1, z1)))
        NODES_DICT[(x1, y1, z1)] = S1

        # S2
        x2 = round(cellcenter[0] + self.__stepx/2, 4)
        y2 = round(cellcenter[1] - (3**0.5)/2 * self.__stepy, 4)
        z2 = round(cellcenter[2] - self.__heightstar, 4)
        S2 = NODES_DICT.get((x2, y2, z2), MACNode(len(NODES_DICT) + 1, (x2, y2, z2)))
        NODES_DICT[(x2, y2, z2)] = S2

        # S3
        x3 = round(cellcenter[0] - self.__stepx/2, 4)
        y3 = round(cellcenter[1] - (3**0.5)/2 * self.__stepy, 4)
        z3 = round(cellcenter[2] + self.__heightstar, 4)
        S3 = NODES_DICT.get((x3, y3, z3), MACNode(len(NODES_DICT) + 1, (x3, y3, z3)))
        NODES_DICT[(x3, y3, z3)] = S3

        # S4
        x4 = round(cellcenter[0] - self.__stepx, 4)
        y4 = round(cellcenter[1], 4)
        z4 = round(cellcenter[2] - self.__heightstar, 4)
        S4 = NODES_DICT.get((x4, y4, z4), MACNode(len(NODES_DICT) + 1, (x4, y4, z4)))
        NODES_DICT[(x4, y4, z4)] = S4

        # S5
        x5 = round(cellcenter[0] - self.__stepx/2, 4)
        y5 = round(cellcenter[1] + (3**0.5)/2 * self.__stepy, 4)
        z5 = round(cellcenter[2] + self.__heightstar, 4)
        S5 = NODES_DICT.get((x5, y5, z5), MACNode(len(NODES_DICT) + 1, (x5, y5, z5)))
        NODES_DICT[(x5, y5, z5)] = S5

        # S6
        x6 = round(cellcenter[0] + self.__stepx/2, 4)
        y6 = round(cellcenter[1] + (3**0.5)/2 * self.__stepy, 4)
        z6 = round(cellcenter[2] - self.__heightstar, 4)
        S6 = NODES_DICT.get((x6, y6, z6), MACNode(len(NODES_DICT) + 1, (x6, y6, z6)))
        NODES_DICT[(x6, y6, z6)] = S6

        # B1
        x = round((2 - self.__dstar)*cellcenter[0] - (1-self.__dstar)*(x6+x5)/2, 4)
        y = round((2 - self.__dstar)*cellcenter[1] - (1-self.__dstar)*(y6+y5)/2, 4)
        z = round((2 - self.__dstar)*cellcenter[2] - (1-self.__dstar)*(z6+z5)/2, 4)
        B1 = NODES_DICT.get((x, y, z), MACNode(len(NODES_DICT) + 1, (x, y, z)))
        NODES_DICT[(x, y, z)] = B1

        # B2
        x = round((2 - self.__dstar)*cellcenter[0] - (1-self.__dstar)*(x1+x6)/2, 4)
        y = round((2 - self.__dstar)*cellcenter[1] - (1-self.__dstar)*(y1+y6)/2, 4)
        z = round((2 - self.__dstar)*cellcenter[2] - (1-self.__dstar)*(z1+z6)/2, 4)
        B2 = NODES_DICT.get((x, y, z), MACNode(len(NODES_DICT) + 1, (x, y, z)))
        NODES_DICT[(x, y, z)] = B2

        # B3
        x = round((2 - self.__dstar)*cellcenter[0] - (1-self.__dstar)*(x1+x2)/2, 4)
        y = round((2 - self.__dstar)*cellcenter[1] - (1-self.__dstar)*(y1+y2)/2, 4)
        z = round((2 - self.__dstar)*cellcenter[2] - (1-self.__dstar)*(z1+z2)/2, 4)
        B3 = NODES_DICT.get((x, y, z), MACNode(len(NODES_DICT) + 1, (x, y, z)))
        NODES_DICT[(x, y, z)] = B3

        # B4
        x = round((2 - self.__dstar)*cellcenter[0] - (1-self.__dstar)*(x2+x3)/2, 4)
        y = round((2 - self.__dstar)*cellcenter[1] - (1-self.__dstar)*(y2+y3)/2, 4)
        z = round((2 - self.__dstar)*cellcenter[2] - (1-self.__dstar)*(z2+z3)/2, 4)
        B4 = NODES_DICT.get((x, y, z), MACNode(len(NODES_DICT) + 1, (x, y, z)))
        NODES_DICT[(x, y, z)] = B4

        # B5
        x = round((2 - self.__dstar)*cellcenter[0] - (1-self.__dstar)*(x4+x3)/2, 4)
        y = round((2 - self.__dstar)*cellcenter[1] - (1-self.__dstar)*(y4+y3)/2, 4)
        z = round((2 - self.__dstar)*cellcenter[2] - (1-self.__dstar)*(z4+z3)/2, 4)
        B5 = NODES_DICT.get((x, y, z), MACNode(len(NODES_DICT) + 1, (x, y, z)))
        NODES_DICT[(x, y, z)] = B5

        # B6
        x = round((2 - self.__dstar)*cellcenter[0] - (1-self.__dstar)*(x4+x5)/2, 4)
        y = round((2 - self.__dstar)*cellcenter[1] - (1-self.__dstar)*(y4+y5)/2, 4)
        z = round((2 - self.__dstar)*cellcenter[2] - (1-self.__dstar)*(z4+z5)/2, 4)
        B6 = NODES_DICT.get((x, y, z), MACNode(len(NODES_DICT) + 1, (x, y, z)))
        NODES_DICT[(x, y, z)] = B6

        # DN1, DN2, DN3, UP1, UP2, UP3
        x = round(cellcenter[0] + 0.5*self.__stepx, 4)
        y = round(cellcenter[1] - (3**0.5)/2 * self.__stepy, 4)
        z = round(cellcenter[2] - self.__hprisma - self.__heightstar, 4)
        DN1 = NODES_DICT.get((x, y, z), MACNode(len(NODES_DICT) + 1, (x, y, z)))
        NODES_DICT[(x, y, z)] = DN1

        x = round(cellcenter[0] - self.__stepx, 4)
        y = round(cellcenter[1], 4)
        z = round(cellcenter[2] - self.__hprisma - self.__heightstar, 4)
        DN2 = NODES_DICT.get((x, y, z), MACNode(len(NODES_DICT) + 1, (x, y, z)))
        NODES_DICT[(x, y, z)] = DN2

        x = round(cellcenter[0] + 0.5 * self.__stepx, 4)
        y = round(cellcenter[1] + (3**0.5) / 2 * self.__stepy, 4)
        z = round(cellcenter[2] - self.__hprisma - self.__heightstar, 4)
        DN3 = NODES_DICT.get((x, y, z), MACNode(len(NODES_DICT) + 1, (x, y, z)))
        NODES_DICT[(x, y, z)] = DN3

        x = round(cellcenter[0] + self.__stepx, 4)
        y = round(cellcenter[1], 4)
        z = round(cellcenter[2] + self.__hprisma + self.__heightstar, 4)
        UP1 = NODES_DICT.get((x, y, z), MACNode(len(NODES_DICT) + 1, (x, y, z)))
        NODES_DICT[(x, y, z)] = UP1

        x = round(cellcenter[0] - 0.5 * self.__stepx, 4)
        y = round(cellcenter[1] - (3**0.5) / 2 * self.__stepy, 4)
        z = round(cellcenter[2] + self.__hprisma + self.__heightstar, 4)
        UP2 = NODES_DICT.get((x, y, z), MACNode(len(NODES_DICT) + 1, (x, y, z)))
        NODES_DICT[(x, y, z)] = UP2

        x = round(cellcenter[0] - 0.5 * self.__stepx, 4)
        y = round(cellcenter[1] + (3**0.5) / 2 * self.__stepy, 4)
        z = round(cellcenter[2] + self.__hprisma + self.__heightstar, 4)
        UP3 = NODES_DICT.get((x, y, z), MACNode(len(NODES_DICT) + 1, (x, y, z)))
        NODES_DICT[(x, y, z)] = UP3

        if self.__nelem == 1:

            nodes = (M1, M2, B1, B2, B3, B4, B5, B6, S1, S2, S3, S4, S5, S6, DN1, DN2, DN3, UP1, UP2, UP3)

            if M1.Coords == M2.Coords:
                sys.exit("M1 es coincidente con M2 porque v1 = 0. Usar el metodo de 4 vigas")

            # CBEAMS that connect the nodes and build the cell
            elemcount = len(ELEMENTS_SET)

            # M1_B1, M1_B2, M1_B3, M1_B4, M1_B5, M1_B6
            M1_B1 = MACElement(elemcount + 1, "CBEAM", [M1, B1], [mat], [prop], vvector)
            ELEMENTS_SET.add(M1_B1)

            M1_B2 = MACElement(elemcount + 2, "CBEAM", [M1, B2], [mat], [prop], vvector)
            ELEMENTS_SET.add(M1_B2)

            M1_B3 = MACElement(elemcount + 3, "CBEAM", [M1, B3], [mat], [prop], vvector)
            ELEMENTS_SET.add(M1_B3)

            M1_B4 = MACElement(elemcount + 4, "CBEAM", [M1, B4], [mat], [prop], vvector)
            ELEMENTS_SET.add(M1_B4)

            M1_B5 = MACElement(elemcount + 5, "CBEAM", [M1, B5], [mat], [prop], vvector)
            ELEMENTS_SET.add(M1_B5)

            M1_B6 = MACElement(elemcount + 6, "CBEAM", [M1, B6], [mat], [prop], vvector)
            ELEMENTS_SET.add(M1_B6)

            # M2_B1, M2_B2, M2_B3, M2_B4, M2_B5, M2_B6
            M2_B1 = MACElement(elemcount + 7, "CBEAM", [M2, B1], [mat], [prop], vvector)
            ELEMENTS_SET.add(M2_B1)

            M2_B2 = MACElement(elemcount + 8, "CBEAM", [M2, B2], [mat], [prop], vvector)
            ELEMENTS_SET.add(M2_B2)

            M2_B3 = MACElement(elemcount + 9, "CBEAM", [M2, B3], [mat], [prop], vvector)
            ELEMENTS_SET.add(M2_B3)

            M2_B4 = MACElement(elemcount + 10, "CBEAM", [M2, B4], [mat], [prop], vvector)
            ELEMENTS_SET.add(M2_B4)

            M2_B5 = MACElement(elemcount + 11, "CBEAM", [M2, B5], [mat], [prop], vvector)
            ELEMENTS_SET.add(M2_B5)

            M2_B6 = MACElement(elemcount + 12, "CBEAM", [M2, B6], [mat], [prop], vvector)
            ELEMENTS_SET.add(M2_B6)

            # S1_B6, B6_S2, S2_B1, B1_S3, S3_B2, B2_S4, S4_B3, B3_S5, S5_B4, B4_S6, S6_B5, B5_S1
            S1_B6 = MACElement(elemcount + 13, "CBEAM", [S1, B6], [mat], [prop], vvector)
            ELEMENTS_SET.add(S1_B6)

            B6_S2 = MACElement(elemcount + 14, "CBEAM", [B6, S2], [mat], [prop], vvector)
            ELEMENTS_SET.add(B6_S2)

            S2_B1 = MACElement(elemcount + 15, "CBEAM", [S2, B1], [mat], [prop], vvector)
            ELEMENTS_SET.add(S2_B1)

            B1_S3 = MACElement(elemcount + 16, "CBEAM", [B1, S3], [mat], [prop], vvector)
            ELEMENTS_SET.add(B1_S3)

            S3_B2 = MACElement(elemcount + 17, "CBEAM", [S3, B2], [mat], [prop], vvector)
            ELEMENTS_SET.add(S3_B2)

            B2_S4 = MACElement(elemcount + 18, "CBEAM", [B2, S4], [mat], [prop], vvector)
            ELEMENTS_SET.add(B2_S4)

            S4_B3 = MACElement(elemcount + 19, "CBEAM", [S4, B3], [mat], [prop], vvector)
            ELEMENTS_SET.add(S4_B3)

            B3_S5 = MACElement(elemcount + 20, "CBEAM", [B3, S5], [mat], [prop], vvector)
            ELEMENTS_SET.add(B3_S5)

            S5_B4 = MACElement(elemcount + 21, "CBEAM", [S5, B4], [mat], [prop], vvector)
            ELEMENTS_SET.add(S5_B4)

            B4_S6 = MACElement(elemcount + 22, "CBEAM", [B4, S6], [mat], [prop], vvector)
            ELEMENTS_SET.add(B4_S6)

            S6_B5 = MACElement(elemcount + 23, "CBEAM", [S6, B5], [mat], [prop], vvector)
            ELEMENTS_SET.add(S6_B5)

            B5_S1 = MACElement(elemcount + 24, "CBEAM", [B5, S1], [mat], [prop], vvector)
            ELEMENTS_SET.add(B5_S1)

            # M1_S1, M1_S3, M1_S5, M2_S2, M2_S4, M2_S6
            M1_S1 = MACElement(elemcount + 25, "CBEAM", [M1, S1], [mat], [prop], vvector)
            ELEMENTS_SET.add(M1_S1)

            M1_S3 = MACElement(elemcount + 26, "CBEAM", [M1, S3], [mat], [prop], vvector)
            ELEMENTS_SET.add(M1_S3)

            M1_S5 = MACElement(elemcount + 27, "CBEAM", [M1, S5], [mat], [prop], vvector)
            ELEMENTS_SET.add(M1_S5)

            M2_S2 = MACElement(elemcount + 28, "CBEAM", [M2, S2], [mat], [prop], vvector)
            ELEMENTS_SET.add(M2_S2)

            M2_S4 = MACElement(elemcount + 29, "CBEAM", [M2, S4], [mat], [prop], vvector)
            ELEMENTS_SET.add(M2_S4)

            M2_S6 = MACElement(elemcount + 30, "CBEAM", [M2, S6], [mat], [prop], vvector)
            ELEMENTS_SET.add(M2_S6)

            # M1_UP1, M1_UP2, M1_UP3, M2_DN1, M2_DN2, M2_DN3
            M1_UP1 = MACElement(elemcount + 31, "CBEAM", [M1, UP1], [mat], [prop], vvector)
            ELEMENTS_SET.add(M1_UP1)

            M1_UP2 = MACElement(elemcount + 32, "CBEAM", [M1, UP2], [mat], [prop], vvector)
            ELEMENTS_SET.add(M1_UP2)

            M1_UP3 = MACElement(elemcount + 33, "CBEAM", [M1, UP3], [mat], [prop], vvector)
            ELEMENTS_SET.add(M1_UP3)

            M2_DN1 = MACElement(elemcount + 34, "CBEAM", [M2, DN1], [mat], [prop], vvector)
            ELEMENTS_SET.add(M2_DN1)

            M2_DN2 = MACElement(elemcount + 35, "CBEAM", [M2, DN2], [mat], [prop], vvector)
            ELEMENTS_SET.add(M2_DN2)

            M2_DN3 = MACElement(elemcount + 36, "CBEAM", [M2, DN3], [mat], [prop], vvector)
            ELEMENTS_SET.add(M2_DN3)

            # All elements in a tuple
            elements = (M1_B1, M1_B2, M1_B3, M1_B4, M1_B5, M1_B6, M2_B1, M2_B2, M2_B3, M2_B4, M2_B5, M2_B6, S1_B6, B6_S2,
                        S2_B1, B1_S3, S3_B2, B2_S4, S4_B3, B3_S5, S5_B4, B4_S6, S6_B5, B5_S1, M1_S1, M1_S3, M1_S5, M2_S2,
                        M2_S4, M2_S6, M1_UP1, M1_UP2, M1_UP3, M2_DN1, M2_DN2, M2_DN3)

        elif self.__nelem == 4:
            internodes = list()
            elements = list()
            elemcount = len(ELEMENTS_SET)

            # M1_B1 intermediate nodes
            for coord in _gen3p(M1.Coords, B1.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M1, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], B1], [mat], [prop], vvector))

            # M1_B2 intermediate nodes
            for coord in _gen3p(M1.Coords, B2.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M1, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], B2], [mat], [prop], vvector))

            # M1_B3 intermediate nodes
            for coord in _gen3p(M1.Coords, B3.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M1, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], B3], [mat], [prop], vvector))

            # M1_B4 intermediate nodes
            for coord in _gen3p(M1.Coords, B4.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M1, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], B4], [mat], [prop], vvector))

            # M1_B5 intermediate nodes
            for coord in _gen3p(M1.Coords, B5.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M1, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], B5], [mat], [prop], vvector))

            # M1_B6 intermediate nodes
            for coord in _gen3p(M1.Coords, B6.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M1, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], B6], [mat], [prop], vvector))

            # Esto se añade porque M1 coincide con M2 si v1 = 0
            if M1.Coords != M2.Coords:

                # M2_B1 intermediate nodes
                for coord in _gen3p(M2.Coords, B1.Coords):
                    internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                    NODES_DICT[coord] = internodes[-1]
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [M2, internodes[-3]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-3], internodes[-2]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-2], internodes[-1]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-1], B1], [mat], [prop], vvector))

                # M2_B2 intermediate nodes
                for coord in _gen3p(M2.Coords, B2.Coords):
                    internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                    NODES_DICT[coord] = internodes[-1]
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [M2, internodes[-3]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-3], internodes[-2]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-2], internodes[-1]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-1], B2], [mat], [prop], vvector))

                # M2_B3 intermediate nodes
                for coord in _gen3p(M2.Coords, B3.Coords):
                    internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                    NODES_DICT[coord] = internodes[-1]
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [M2, internodes[-3]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-3], internodes[-2]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-2], internodes[-1]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-1], B3], [mat], [prop], vvector))

                # M2_B4 intermediate nodes
                for coord in _gen3p(M2.Coords, B4.Coords):
                    internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                    NODES_DICT[coord] = internodes[-1]
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [M2, internodes[-3]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-3], internodes[-2]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-2], internodes[-1]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-1], B4], [mat], [prop], vvector))

                # M2_B5 intermediate nodes
                for coord in _gen3p(M2.Coords, B5.Coords):
                    internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                    NODES_DICT[coord] = internodes[-1]
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [M2, internodes[-3]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-3], internodes[-2]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-2], internodes[-1]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-1], B5], [mat], [prop], vvector))

                # M2_B6 intermediate nodes
                for coord in _gen3p(M2.Coords, B6.Coords):
                    internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                    NODES_DICT[coord] = internodes[-1]
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [M2, internodes[-3]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-3], internodes[-2]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-2], internodes[-1]], [mat], [prop], vvector))
                elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                           [internodes[-1], B6], [mat], [prop], vvector))

            # S1_B6 intermediate nodes
            for coord in _gen3p(S1.Coords, B6.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [S1, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], B6], [mat], [prop], vvector))

            # B6_S2 intermediate nodes
            for coord in _gen3p(B6.Coords, S2.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [B6, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], S2], [mat], [prop], vvector))

            # S2_B1 intermediate nodes
            for coord in _gen3p(S2.Coords, B1.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [S2, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], B1], [mat], [prop], vvector))

            # B1_S3 intermediate nodes
            for coord in _gen3p(B1.Coords, S3.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [B1, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], S3], [mat], [prop], vvector))

            # S3_B2 intermediate nodes
            for coord in _gen3p(S3.Coords, B2.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [S3, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], B2], [mat], [prop], vvector))

            # B2_S4 intermediate nodes
            for coord in _gen3p(B2.Coords, S4.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [B2, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], S4], [mat], [prop], vvector))

            # S4_B3 intermediate nodes
            for coord in _gen3p(S4.Coords, B3.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [S4, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], B3], [mat], [prop], vvector))

            # B3_S5 intermediate nodes
            for coord in _gen3p(B3.Coords, S5.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [B3, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], S5], [mat], [prop], vvector))

            # S5_B4 intermediate nodes
            for coord in _gen3p(S5.Coords, B4.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [S5, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], B4], [mat], [prop], vvector))

            # B4_S6 intermediate nodes
            for coord in _gen3p(B4.Coords, S6.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [B4, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], S6], [mat], [prop], vvector))

            # S6_B5 intermediate nodes
            for coord in _gen3p(S6.Coords, B5.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [S6, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], B5], [mat], [prop], vvector))

            # B5_S1 intermediate nodes
            for coord in _gen3p(B5.Coords, S1.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [B5, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], S1], [mat], [prop], vvector))

            # M1_S1 intermediate nodes
            for coord in _gen3p(M1.Coords, S1.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M1, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], S1], [mat], [prop], vvector))

            # M1_S3 intermediate nodes
            for coord in _gen3p(M1.Coords, S3.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M1, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], S3], [mat], [prop], vvector))

            # M1_S5 intermediate nodes
            for coord in _gen3p(M1.Coords, S5.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M1, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], S5], [mat], [prop], vvector))

            # M2_S2 intermediate nodes
            for coord in _gen3p(M2.Coords, S2.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M2, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], S2], [mat], [prop], vvector))

            # M2_S4 intermediate nodes
            for coord in _gen3p(M2.Coords, S4.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M2, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], S4], [mat], [prop], vvector))

            # M2_S6 intermediate nodes
            for coord in _gen3p(M2.Coords, S6.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M2, internodes[-3]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], S6], [mat], [prop], vvector))

            # M1_UP1 intermediate nodes
            for coord in _gen3p(M1.Coords, UP1.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M1, internodes[-3]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], UP1], [mat], [prop], vvector, aux=True))

            # M1_UP2 intermediate nodes
            for coord in _gen3p(M1.Coords, UP2.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M1, internodes[-3]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], UP2], [mat], [prop], vvector, aux=True))

            # M1_UP3 intermediate nodes
            for coord in _gen3p(M1.Coords, UP3.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M1, internodes[-3]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], UP3], [mat], [prop], vvector, aux=True))

            # M2_DN1 intermediate nodes
            for coord in _gen3p(M2.Coords, DN1.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M2, internodes[-3]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-3], internodes[-2]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-2], internodes[-1]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [internodes[-1], DN1], [mat], [prop], vvector, aux=True))

            # M2_DN2 intermediate nodes
            for coord in _gen3p(M2.Coords, DN2.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                       [M2, internodes[-3]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                        [internodes[-3], internodes[-2]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                        [internodes[-2], internodes[-1]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                        [internodes[-1], DN2], [mat], [prop], vvector, aux=True))

            # M2_DN3 intermediate nodes
            for coord in _gen3p(M2.Coords, DN3.Coords):
                internodes.append(MACNode(len(NODES_DICT) + 1, coord))
                NODES_DICT[coord] = internodes[-1]
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                        [M2, internodes[-3]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                        [internodes[-3], internodes[-2]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                        [internodes[-2], internodes[-1]], [mat], [prop], vvector, aux=True))
            elements.append(MACElement(elemcount + len(elements) + 1, "CBEAM",
                                        [internodes[-1], DN3], [mat], [prop], vvector, aux=True))

            for element in elements:
                ELEMENTS_SET.add(element)

            elements = tuple(elements)
            nodes = tuple([M1, M2, B1, B2, B3, B4, B5, B6, S1, S2, S3, S4, S5, S6, DN1, DN2, DN3, UP1, UP2, UP3]
                          + internodes)


        return nodes, elements


def _gen3p(point1: tuple, point2: tuple) -> tuple:
    x1, y1, z1 = point1
    x2, y2, z2 = point2

    # Calculate the intermediate points
    x_mid = (x1 + x2) / 2
    y_mid = (y1 + y2) / 2
    z_mid = (z1 + z2) / 2

    x_quarter = (3 * x1 + x2) / 4
    y_quarter = (3 * y1 + y2) / 4
    z_quarter = (3 * z1 + z2) / 4

    x_three_quarters = (x1 + 3 * x2) / 4
    y_three_quarters = (y1 + 3 * y2) / 4
    z_three_quarters = (z1 + 3 * z2) / 4

    # Create and return the three points between
    point_mid = (x_mid, y_mid, z_mid)
    point_quarter = (x_quarter, y_quarter, z_quarter)
    point_three_quarters = (x_three_quarters, y_three_quarters, z_three_quarters)

    return point_quarter, point_mid, point_three_quarters



def set_structure(**kwargs) -> MACAuxetic:
    """ Function to set the structure of the cell. It uses **kwargs, a dictionary of parameters. Depending on the
    type of the structure an object of the corresponding class is created and returned. Types supported are:
        - Auxetic: set_structure(type="Auxetic", djoint=float, dstar=float, heightstar=float, hcapas=float,+
                                 hprisma=float, stepx=0.1, stepy=0.1, nelem=1)
    """
    if kwargs["type"] == "Auxetic":

        structure = MACAuxetic(djoint=kwargs["djoint"], dstar=kwargs["dstar"], heightstar=kwargs["heightstar"],
                               hcapas=kwargs["hcapas"], hprisma=kwargs["hprisma"], stepx=kwargs["stepx"],
                               stepy=kwargs["stepy"], nelem=kwargs["nelem"])

        return structure



# Copyright (c) 2022-2023 MetamaterialAnalysisCode

from MetamaterialAnalysisCode.classes.MACAnalysis import set_analysis
from MetamaterialAnalysisCode.classes.MACModel import set_model
from MetamaterialAnalysisCode.classes.MACTable import set_table
from MetamaterialAnalysisCode.classes.MACMaterial import set_material
from MetamaterialAnalysisCode.classes.MACProperty import set_property
from MetamaterialAnalysisCode.classes.MACStructure import set_structure
from MetamaterialAnalysisCode.classes.MACLoadCase import set_load, set_constraint, set_eigr
from MetamaterialAnalysisCode.classes.MACSubcase import set_subcase
from MetamaterialAnalysisCode.modules.MACRun import run_optistruct
from MetamaterialAnalysisCode.classes.MACNonLinear import set_nlparmld, set_nlout


"""
Element module for the Metamaterial Analysis Code (MAC). It represents a single element.
"""

from .MACNode import MACNode
from .MACProperty import MACProperty
from .MACMaterial import MACMaterial

class MACElement:
    """
    Element class for the Metamaterial Analysis Code (MAC). It represents a single element.

    Attributes:
        ID: element ID
        Type: type of the element (CBEAM, CQUAD, etc)
        Nodes: list of nodes that compose the element
        Material: list of the materials of the element
        Properties: list of the properties of the element
        Vvector: vector that determines the orientation of the element

    """

    def __init__(self, id_: int, type_: str, nodes: list[MACNode], material: list[MACMaterial],
                 properties: list[MACProperty], vvector: tuple, aux: bool = False):
        """
        Constructor for MACElement class
        """
        self.__id = id_
        self.__type = type_
        self.__nodes = nodes
        self.__material = material
        self.__properties = properties
        self.__vvector = vvector
        self.Aux = aux  # Propiedad auxiliar para saber si es de los elementos que se pueden borrar

    @property
    def ID(self):
        return self.__id

    @property
    def Type(self):
        return self.__type

    @property
    def Nodes(self):
        return self.__nodes

    @property
    def Materials(self):
        return self.__material

    @property
    def Property(self):
        return self.__properties

    @property
    def Vvector(self):
        return self.__vvector

    # Method to print a node. It uses the 8 characters format of Optistruct.
    def __str__(self):

        match self.Type:

            case "CBEAM":

                idspaces = " " * (8 - len(str(self.ID)))
                propspaces = " " * (8 - len(str(self.Property[0].ID)))
                node0spaces = " " * (8 - len(str(self.Nodes[0].ID)))
                node1spaces = " " * (8 - len(str(self.Nodes[1].ID)))
                vv0spaces = " " * (8 - len(str(self.Vvector[0])))
                vv1spaces = " " * (8 - len(str(self.Vvector[1])))
                vv2spaces = " " * (8 - len(str(self.Vvector[2])))

                return f"CBEAM   {idspaces}{self.ID}{propspaces}{self.Property[0].ID}{node0spaces}{self.Nodes[0].ID}" + \
                       f"{node1spaces}{self.Nodes[1].ID}{vv0spaces}{self.Vvector[0]}{vv1spaces}{self.Vvector[1]}" + \
                       f"{vv2spaces}{self.Vvector[2]}\n"

            case "CQUAD":

                return "CQUAD is not yet implemented"

    def __hash__(self):
        return hash(self.ID)


"""
Module for running the .fem files in optistruct
"""
from subprocess import run
import os
from time import sleep
from shutil import move

def run_optistruct(path: str, numcores: int) -> None:
    """
    Function that runs the .fem file using optistruct. It should be the same as given for MACAnalysis.write()
    The number of cores is given by the user. Don't surpass the number of cores of your computer.
    """
    # Code that runs the .fem file using optistruct
    run([r"C:\Program Files\Altair\2022\hwsolvers\scripts\optistruct.bat", path, "-nt", str(numcores)])


def secuencial_run_optistruct(directory: str, numcores: int) -> None:
    """
    Function that runs the .fem files in a directory using optistruct.
    The number of cores is given by the user. Don't surpass the number of cores of your computer.
    """

    fem_files = {file for file in os.listdir(directory) if file.endswith(".fem")}
    for fem_file in fem_files:

        # Creo una nueva carpeta con el nombre del archivo .fem
        newdirectory = os.path.join(directory, fem_file.split(".")[0])
        os.makedirs(newdirectory)

        # Muevo el archivo .fem a la nueva carpeta
        move(os.path.join(directory, fem_file), newdirectory)

        # Ejecuto el archivo .fem
        run_optistruct(os.path.join(newdirectory, fem_file), numcores)

        # Espero a que se cree el archivo .h3d
        while not fem_file.split(".")[0] + ".h3d" in os.listdir(directory):
            sleep(5)
            for runing_file in os.listdir(directory):
                if not runing_file.endswith(".txt") and not runing_file.endswith(".message"):
                    break
        # Una vez creado, paso al siguiente archivo


# path where this code is executed
path = os.path.dirname(os.path.abspath(__file__))

secuencial_run_optistruct(path, 20)




"""
Material module for Metamaterial Analysis Code (MAC). It represents a single material.
"""
from typing import Any
from abc import ABC, abstractmethod

from MetamaterialAnalysisCode.classes.MACTable import MACTable

class MACMaterial(ABC):
    """
    Material base class for Metamaterial Analysis Code (MAC). The materials classes (MACMatNL or MACMat) are based
    on this base class

    Attributes:
        ID: material ID
        Type: type of the material (Supported: MATS1)
    """
    @abstractmethod
    def __init__(self, id_: int, type_: str):
        """
        Constructor for MACMaterial class
        """
        self.ID = id_
        self.Type = type_


class MACMatNL(MACMaterial):
    """
    Non-Linear Material class for Metamaterial Analysis Code (MAC). It represents a single material.

    Attributes:
        ID: material ID
        StressStrain: stress-strain curve of the material
        Type: type of the material (Supported: MATS1)
        NonLinearity: type of non-linearity of the material (LINEAR, PLASTIC, NLELAST)
        YieldStress: Stress at the yield point in the strain-stress curve of the material
    """

    def __init__(self, id_: int, type_: str, stressstrain: MACTable = None, nonlinearity: str = "PLASTIC",
                 yieldstress: float = 0.0):
        """
        Constructor for MACMaterial class
        """
        super().__init__(id_, type_)
        self.StressStrain = stressstrain
        self.NonLinearity = nonlinearity
        self.YieldStress = yieldstress

    # Method to print a node. It uses the 8 characters format of Optistruct.
    def __str__(self):

        idspaces = " " * (8 - len(str(self.ID)))
        stressspaces = " " * (8 - len(str(self.StressStrain.ID)))
        nonlinearityspaces = " " * (8 - len(self.NonLinearity))

        return f"MATS1   {self.ID}{idspaces}{self.StressStrain.ID}{stressspaces}{self.NonLinearity}" + \
               f"{nonlinearityspaces}\n" + "\n" + " " * 8 + f"{self.YieldStress}\n"


class MACMat(MACMaterial):
    """
    Linear isotropic Material class for Metamaterial Analysis Code (MAC). It represents a single material.

    Attributes:
        ID: material ID
        Type: MAT1
        E: Young's modulus
        G: Shear modulus
        Nu: Poisson's ratio
        Rho: Density
        TExp: Thermal expansion coefficient
        TRef: Reference temperature
        GE: Structural damping coefficient
        ST: Stress limit for tension
        SC: Stress limit for compression
        SS: Stress limit for shear
    """

    def __init__(self, id_: int, type_: str, e: float, nu: float, g: float = None, rho: float = None,
                 texp: float = None, tref: float = None, ge: float = None,
                 st: float = None, sc: float = None, ss: float = None):
        """
        Constructor for MACMat class
        """
        super().__init__(id_, type_)
        self.E = e
        self.G = g
        self.Nu = nu
        self.Rho = rho
        self.TExp = texp
        self.TRef = tref
        self.GE = ge
        self.ST = st
        self.SC = sc
        self.SS = ss

    def __str__(self):

        idspaces = " " * (8 - len(str(self.ID)))
        e = "{:.5f}".format(self.E)[:7]
        nu = "{:.5f}".format(self.Nu)[:7]

        if self.G is None:
            g = " "*7
        else:
            g = "{:.5f}".format(self.G)[:7]

        if self.Rho is None:
            rho = " "*7
        else:
            rho = "{:.5f}".format(self.Rho)[:7]

        if self.TExp is None:
            texp = " "*7
        else:
            texp = "{:.5f}".format(self.TExp)[:7]

        if self.TRef is None:
            tref = " "*7
        else:
            tref = "{:.5f}".format(self.TRef)[:7]

        if self.GE is None:
            ge = " "*7
        else:
            ge = "{:.5f}".format(self.GE)[:7]

        if self.ST is None:
            st = " "*7
        else:
            st = "{:.5f}".format(self.ST)[:7]

        if self.SC is None:
            sc = " "*7
        else:
            sc = "{:.5f}".format(self.SC)[:7]

        if self.SS is None:
            ss = " "*7
        else:
            ss = "{:.5f}".format(self.SS)[:7]

        if (self.ST is not None) or (self.SC is not None) or (self.SS is not None):

            return f"MAT1    {self.ID}{idspaces} {e} {g} {nu} {rho} {texp} {tref} {ge}\n" + \
                   f"         {st} {sc} {ss}\n"
        else:

            return f"MAT1    {self.ID}{idspaces} {e} {g} {nu} {rho} {texp} {tref} {ge}\n"


def set_material(**kwargs: dict[str: Any]) -> MACMat | MACMatNL:
    """
    Function to set a material as MACMaterial object. The function use the kwargs dictionary. The supported type of
    materials are:
        - MATS1: set_material(id=int, type="MATS1", stressstrain=MACTable, nonlinearity="NLELAST"|"PLASTIC  ", yieldstress=float)
    """
    if kwargs["type"] == "MATS1":

        stressstrain = kwargs["stressstrain"]
        type_ = kwargs["type"]
        nonlinearity = kwargs["nonlinearity"]
        yieldstress = kwargs["yieldstress"]
        id_ = kwargs["id"]
        material = MACMatNL(id_, type_, stressstrain, nonlinearity, yieldstress)

        return material

    elif kwargs["type"] == "MAT1":

        id_ = kwargs["id"]
        type_ = kwargs["type"]
        e = kwargs["e"]
        nu = kwargs["nu"]
        g = kwargs.get("g", None)
        rho = kwargs.get("rho", None)
        texp = kwargs.get("texp", None)
        tref = kwargs.get("tref", None)
        ge = kwargs.get("ge", None)
        st = kwargs.get("st", None)
        sc = kwargs.get("sc", None)
        ss = kwargs.get("ss", None)
        material = MACMat(id_, type_, e, nu, g, rho, texp, tref, ge, st, sc, ss)

        return material

    else:
        raise ValueError("The material type is not supported by the MACMaterial class.")


"""
Module with the MACLoadCase base class and its derived classes.
"""
from MetamaterialAnalysisCode.classes.MACNode import MACNode


class MACLoadCase:
    """
    Base class for the load cases. Supported classes: FORCE, SPC, EIGRL.

    Attributes:
        ID: load case ID
        Type: load case type
    """

    def __init__(self, id_: int, type_: str):
        """
        Constructor for MACLoadCase class
        """
        self.ID = id_
        self.Type = type_


class MACForce(MACLoadCase):
    """
    Class for the FORCE load case. It derives from MACLoadCase.

    Attributes:
        ID: load case ID
        Type: load case type
        Node/Set: node where the force is applied/Set of nodes where the force is applied
        Direction: direction of the force
        Magnitude: magnitude of the force
        GSet: bool that is True if the force is applied to a set of nodes, False otherwise
    """

    def __init__(self, id_: int, node: list[MACNode, ...], direction: tuple[float, float, float], magnitude: float):
        """
        Constructor for MACForce class
        """
        super().__init__(id_, "FORCE")
        self.Nodes = node
        self.Direction = direction
        self.Magnitude = magnitude
        if len(node) > 1:
            self.__gset = True
        else:
            self.__gset = False

    def __str__(self):
        """
        Method to print a FORCE load case. It uses the 8 characters format of Optistruct.
        """
        idspaces = " " * (8 - len(str(self.ID)))
        systemspaces = " " * 8

        # 8 characters for each coordinate. It uses scientific notation with 3 decimal places.
        x = "{:.5f}".format(self.Direction[0])[:7]
        y = "{:.5f}".format(self.Direction[1])[:7]
        z = "{:.5f}".format(self.Direction[2])[:7]
        magnitude = "{:.5f}".format(self.Magnitude)[:7]

        if self.GSet:
            idset = str(self.ID*100)
            idspaceset = " " * (8 - len(str(idset)))
            gset = f"SET     {idset}{idspaceset}    GRID    LIST\n" + " "*8
            count = 0

            for node in self.Nodes:
                nodespaces = " " * (8 - len(str(node.ID)))
                gset += f"{nodespaces}{node.ID}"
                count += 1
                if count == 8:
                    gset += "\n" + " "*8
                    count = 0
            gset += "\n"

            return f"FORCE   {idspaces}{self.ID}{idspaceset}{idset}{systemspaces} {magnitude}" + \
                   f" {x} {y} {z}\n" + " "*8 + "GSET" + "\n" + gset
        else:
            nodespaces = " " * (8 - len(str(self.Nodes[0].ID)))
            return f"FORCE   {idspaces}{self.ID}{nodespaces}{self.Nodes[0].ID}{systemspaces}{magnitude}" + \
                   f" {x} {y} {z}\n"

    @property
    def GSet(self) -> bool:
        """
        If True, the force is applied to a set of nodes, if False, the force is applied to a single node.
        """
        return self.__gset


class MACSpc(MACLoadCase):
    """
    Class for the SPC load case. It derives from MACLoadCase.

    Attributes:
        ID: load case ID
        Type: load case type
        Node/Set: node where the force is applied/Set of nodes where the force is applied
        Components: Freedom degrees that are constrained
        Displacement: Value of enforced displacement for all coordinates
        GSet: bool that is True if the force is applied to a set of nodes, False otherwise
    """

    def __init__(self, id_: int, node: list[MACNode, ...], components: list[int, ...], displacement: float,
                 enfdisp: bool = False):
        """
        Constructor for MACSPC class
        """
        super().__init__(id_, "SPC")
        self.Nodes = node
        self.Components = components
        self.Displacement = displacement
        self.EnfDisp = enfdisp  # Enforced Displacement
        if len(node) > 1:
            self.__gset = True
        else:
            self.__gset = False

    @property
    def GSet(self) -> bool:
        """
        If True, the force is applied to a set of nodes, if False, the force is applied to a single node.
        """
        return self.__gset

    def __str__(self):
        """
        Method to print a SPC load case. It uses the 8 characters format of Optistruct.
        """
        idspaces = " " * (8 - len(str(self.ID)))

        # 8 characters for each coordinate. It uses scientific notation with 3 decimal places.
        displacement = "{:.5f}".format(self.Displacement)[:7]
        components = ""
        for comp in self.Components:
            components += str(comp)
        compspaces = " " * (8 - len(components))

        if self.GSet:
            idset = str(self.ID*100+self.Nodes[0].ID)
            idspaceset = " " * (8 - len(str(idset)))
            gset = f"SET     {idspaceset}{idset}    GRID    LIST\n" + " "*8
            count = 0

            for node in self.Nodes:
                nodespaces = " " * (8 - len(str(node.ID)))
                gset += f"{nodespaces}{node.ID}"
                count += 1
                if count == 8:
                    gset += "\n" + " "*8
                    count = 0
            gset += "\n"

            if self.EnfDisp:
                return f"SPCD    {idspaces}{self.ID}{idspaceset}{idset}{compspaces}{components} {displacement}\n" + \
                       " " * 8 + "    GSET" + "\n" + gset

            else:
                return f"SPC     {idspaces}{self.ID}{idspaceset}{idset}{compspaces}{components} {displacement}\n" + \
                   " "*8 + "    GSET" + "\n" + gset

        else:

            nodespaces = " " * (8 - len(str(self.Nodes[0].ID)))

            if self.EnfDisp:
                idspcadd = str(self.ID * 100 + 1)
                spcaddspaces = " " * (8 - len(str(idspcadd)))
                return f"SPCD    {spcaddspaces}{idspcadd}{nodespaces}{self.Nodes[0].ID}{compspaces}{components} " + \
                       f"{displacement}\n"

            else:
                return f"SPC     {self.ID}{idspaces}{nodespaces}{self.Nodes[0].ID}{compspaces}{components} " + \
                       f"{displacement}\n"


class MACEigrl(MACLoadCase):
    """
    Class for the EIGRL load case. It derives from MACLoadCase.

    Attributes:
        ID: load case ID
        Type: load case type
        NumRoots: number of roots to be calculated
    """

    def __init__(self, id_: int, numroots: int):
        """
        Constructor for MACEigrl class
        """
        super().__init__(id_, "EIGRL")
        self.NumRoots = numroots

    def __str__(self):
        """
        Method to print an EIGRL load case. It uses the 8 characters format of Optistruct.
        """
        idspaces = " " * (8 - len(str(self.ID)))
        numrootsspaces = " " * (8 - len(str(self.NumRoots)))
        vibrationspaces = " " * 8 * 2
        return f"EIGRL   {idspaces}{self.ID}{vibrationspaces}{numrootsspaces}{self.NumRoots}\n"


def set_load(**kwargs) -> MACForce | MACSpc:
    """
    Function to set the load case. It could return a MACForce, MACSpc or MACEigrl. Its arguments should be given as
    keyword arguments:
        - MACForce: set_load_case(id=int, type="FORCE" nodes=[MACNode], direction=tuple[float, ...], magnitude=float)
        - MACSpc: set_load_case(id=int, type="SPC" nodes=[MACNode], components=list[int, ...], displacement=float)
        - MACEigrl: set_load_case(id=int, type="EIGRL", numroots=int)
    """

    if kwargs["type"] == "FORCE":
        load_case = MACForce(kwargs["id"], kwargs["nodes"], kwargs["direction"], kwargs["magnitude"])

    elif kwargs["type"] == "SPC":
        load_case = MACSpc(kwargs["id"], kwargs["nodes"], kwargs["components"], kwargs["displacement"],
                           kwargs.get("load", True))

    else:
        raise ValueError("Invalid LOAD type")

    return load_case


def set_constraint(id: int, nodes: list[MACNode, ...], components: list[int, ...], displacement: float,
                   load: bool = False) -> MACSpc:
    """
    Function to set the constraint load case. It returns a MACSpc.
        - MACSpc: set_load_case(id=int, nodes=[MACNode], components=list[int, ...], displacement=float,
                                load=bool)
    """
    return MACSpc(id, nodes, components, displacement, load)


def set_eigr(id: int, type: str, numroots: int) -> MACEigrl:
    """
    Function to set the EIGRL load case. It returns a MACEigrl.
        - MACEigrl: set_load_case(id=int, type="EIGRL", numroots=int)
    """
    if type == "EIGRL":
        return MACEigrl(id, numroots)
    else:
        raise ValueError("Invalid EIGR type")


"""
Node module for the Metamaterial Analysis Code (MAC). It represents a single grid.
"""


class MACNode:
    """
    Node class for the Metamaterial Analysis Code (MAC). It represents a single grid.

    Attributes:
        ID: node ID
        Coords: Coordinates of the node
    """

    def __init__(self, id_: int, coords: tuple[float, ...]):
        """
        Constructor for MACNode class
        """
        self.__id = id_
        self.__coords = coords

    @property
    def ID(self):
        return self.__id

    @property
    def Coords(self) -> tuple:
        return self.__coords

    @Coords.setter
    def Coords(self, value):
        self.__coords = value

    # Method to compare two nodes. It returns True if the nodes are the same (they have the same coordinates),
    # False otherwise.
    def __eq__(self, other) -> bool:
        if isinstance(other, MACNode):
            return self.Coords == other.Coords
        else:
            return False

    # Method to hash a node. It uses the coordinates of the node as hash to be implemented in a set.
    def __hash__(self):
        return hash((self.Coords[0], self.Coords[1], self.Coords[2], self.ID))

    # Method to print a node. It uses the 8 characters format of Optistruct.
    def __str__(self):

        idspaces = " " * (8 - len(str(self.ID)))
        systemspaces = " " * 8

        # 8 characters for each coordinate. It uses scientific notation with 3 decimal places.
        x = "{:.5f}".format(self.Coords[0])[:7]
        y = "{:.5f}".format(self.Coords[1])[:7]
        z = "{:.5f}".format(self.Coords[2])[:7]

        return f"GRID    {idspaces}{self.ID}{systemspaces} {x} {y} {z}\n"



"""
Module with global variables for the Metamaterial Analysis Code (MAC) package.
"""
from typing import Set
from MetamaterialAnalysisCode.classes.MACElement import MACElement

global NODES_DICT
NODES_DICT = dict()

global ELEMENTS_SET
ELEMENTS_SET: Set[MACElement] = set()

MAC_VERSION = "0.0"

AUTHOR = "Manuel Sanchez Garcia"
CONTRIBUTORS = []
